// *************************************************************************************************
//
// Horde3D
//   Next-Generation Graphics Engine
// --------------------------------------
// Copyright (C) 2006-2009 Nicolas Schulz
//
// This software is distributed under the terms of the Eclipse Public License v1.0.
// A copy of the license may be obtained at: http://www.eclipse.org/legal/epl-v10.html
//
// *************************************************************************************************

#define DONT_ADD_HACK

#include "egRendererBaseOpenGLES11.h"
#include "egModules.h"
#include "egCom.h"
#include "utOpenGLES11.h"

#include "utDebug.h"


// =================================================================================================
// GPUTimer
// =================================================================================================

GPUTimer::GPUTimer()
{
	reset();
}


GPUTimer::~GPUTimer()
{
}


void GPUTimer::beginQuery( uint32 frameID )
{
	return;
}


void GPUTimer::endQuery()
{
}


bool GPUTimer::updateResults()
{
	return false;
}


void GPUTimer::reset()
{
}


// =================================================================================================
// RendererBase
// =================================================================================================

RendererBase::RendererBase()
{
	_vpX = 0; _vpY = 0;
	_vpWidth = 320; _vpHeight = 240;
	_curShaderObj = 0;
	_curRendBuf = 0; _outputBufferIndex = 0;
	_textureMem = 0; _bufferMem = 0;
}


RendererBase::~RendererBase()
{
}


void RendererBase::initStates()
{
	glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
	glDisable( GL_MULTISAMPLE );
}


bool RendererBase::init()
{
	bool failed = false;

	char *vendor = (char *)glGetString( GL_VENDOR );
	char *renderer = (char *)glGetString( GL_RENDERER );
	char *version = (char *)glGetString( GL_VERSION );
	
	Modules::log().writeInfo( "Initializing GL1.5 backend using OpenGL driver '%s' by '%s' on '%s'",
	                          version, vendor, renderer );
	
	// Init extensions
	if( !initOpenGLExtensions() )
	{	
		Modules::log().writeError( "Could not find all required OpenGL function entry points" );
		failed = true;
	}

	// Check that OpenGLES 1.1 (or OpenGL 1.5) is available
#ifdef RENDERER_OPENGL15
	if( glExt::majorVersion * 10 + glExt::minorVersion < 15 )
#else
	if( glExt::majorVersion * 10 + glExt::minorVersion < 11 )
#endif
	{
		Modules::log().writeError( "OpenGL-ES 1.1 (or OpenGL1.5) not available" );
		failed = true;
	}
	
	// Check that required extensions are supported
	if( !glExt::OES_framebuffer_object)
	{
		Modules::log().writeError( "Extension OES_framebuffer_object not supported" );
		failed = true;
	}
	if( !glExt::EXT_texture_filter_anisotropic )
	{
		Modules::log().writeError( "Extension EXT_texture_filter_anisotropic not supported" );
//		failed = true;	//no support on mesa
	}
	if( !glExt::EXT_texture_compression_s3tc )
	{
		Modules::log().writeError( "Extension EXT_texture_compression_s3tc not supported" );
	}
	if( !glExt::EXT_texture_compression_pvrtc )
	{
		Modules::log().writeError( "Extension EXT_texture_compression_pvrtc not supported" );
	}
	
	if( failed )
	{
		Modules::log().writeError( "Failed to init renderer backend, debug info following" );
		char *exts = (char *)glGetString( GL_EXTENSIONS );
		Modules::log().writeInfo( "Supported extensions: '%s'", exts );

		return false;
	}
	
	// only 16bit depth is supported in OES_framebuffer_object
	_depthFormat = GL_DEPTH_COMPONENT16_OES;
	glGetIntegerv(GL_FRAMEBUFFER_BINDING_OES, &_defaultFramebuffer);
	
	initStates();

	return true;
}


void RendererBase::resize( int x, int y, int width, int height )
{
	_vpX = x; _vpY = y; _vpWidth = width; _vpHeight = height;
}


// =================================================================================================
// Buffers
// =================================================================================================

uint32 RendererBase::createVertexBuffer( uint32 size, void *data )
{
	RBBuffer buf;

	buf.type = GL_ARRAY_BUFFER;
	buf.size = size;
	glGenBuffers( 1, &buf.glObj );
	glBindBuffer( buf.type, buf.glObj );
	glBufferData( buf.type, size, data, GL_DYNAMIC_DRAW );
	glBindBuffer( buf.type, 0 );

	_bufferMem += size;
	return _buffers.add( buf );
}


uint32 RendererBase::createIndexBuffer( uint32 size, void *data )
{
	RBBuffer buf;

	buf.type = GL_ELEMENT_ARRAY_BUFFER;
	buf.size = size;
	glGenBuffers( 1, &buf.glObj );
	glBindBuffer( buf.type, buf.glObj );
	glBufferData( buf.type, size, data, GL_DYNAMIC_DRAW );
	glBindBuffer( buf.type, 0 );

	_bufferMem += size;
	return _buffers.add( buf );
}


void RendererBase::releaseBuffer( uint32 bufObj )
{
	if( bufObj == 0 ) return;
	
	RBBuffer &buf = _buffers.getRef( bufObj );
	glDeleteBuffers( 1, &buf.glObj );

	_bufferMem -= buf.size;
	_buffers.remove( bufObj );
}


void RendererBase::updateBufferData( uint32 bufObj, uint32 offset, uint32 size, void *data )
{
	const RBBuffer &buf = _buffers.getRef( bufObj );
	ASSERT( offset + size <= buf.size );
	glBindBuffer( buf.type, buf.glObj );
	
	if( offset == 0 &&  size == buf.size )
	{
		// Replacing the whole buffer can help the driver to avoid pipeline stalls
		glBufferData( buf.type, size, data, GL_DYNAMIC_DRAW );
		return;
	}

	glBufferSubData( buf.type, offset, size, data );
	glBindBuffer( buf.type, 0x0 );
}


uint32 RendererBase::cloneBuffer( uint32 bufObj )
{
	Modules::log().writeWarning( "Cloning is unsupported in OpenGL ES 1.1 backend" );
	return 0;	//cloning is not supported, sorry
}


void RendererBase::bindVertexBuffer( uint32 slot, uint32 vbObj, uint32 offset, uint32 stride )
{
	_vertBufSlots[slot].vbObj = vbObj;
	_vertBufSlots[slot].offset = offset;
	_vertBufSlots[slot].stride = stride;
}


void RendererBase::bindIndexBuffer( uint32 bufObj )
{
	if( bufObj == 0 )
	{
		glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, 0 );
	}
	else
	{
		const RBBuffer &buf = _buffers.getRef( bufObj );
		glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, buf.glObj );
	}
}


// =================================================================================================
// Textures
// =================================================================================================

uint32 RendererBase::calcTextureSize( TextureFormats::List format, int width, int height )
{
	switch( format )
	{
	case TextureFormats::BGRA8:
		return width * height * 4;
    case TextureFormats::RGB5_A1:
            return width * height * 2;
    case TextureFormats::RGB565:
            return width * height * 2;
    case TextureFormats::RGBA4:
            return width * height * 2;
	case TextureFormats::DXT1:
		return std::max( width / 4, 1 ) * std::max( height / 4, 1 ) * 8;
	case TextureFormats::DXT3:
		return std::max( width / 4, 1 ) * std::max( height / 4, 1 ) * 16;
	case TextureFormats::DXT5:
		return std::max( width / 4, 1 ) * std::max( height / 4, 1 ) * 16;
	case TextureFormats::RGBA16F:
		return width * height * 8;
	case TextureFormats::RGBA32F:
		return width * height * 16;
    case TextureFormats::PVRTC2:
	case TextureFormats::PVRTCA2:
		return (std::max( width, 16 ) * std::max( height, 8 ) * 2 + 7) / 8;
	case TextureFormats::PVRTC4:
	case TextureFormats::PVRTCA4:
		return (std::max( width, 8 ) * std::max( height, 8 ) * 4 + 7) / 8;
	default:
		return 0;
	}
}


uint32 RendererBase::createTexture( TextureTypes::List type, int width, int height, TextureFormats::List format,
	                                bool hasMips, bool genMips, bool compress, bool sRGB )
{
	if( !_caps[RenderCaps::Tex_NPOT] )
	{
		// Check if texture is NPOT
		if( (width & (width-1)) != 0 || (height & (height-1)) != 0 )
			Modules::log().writeWarning( "Texture has non-power-of-two dimensions although NPOT is not supported by GPU" );
	}

	if (sRGB && Modules::config().sRGBLinearization)
		Modules::log().writeWarning( "sRGBLinearization is not supported" );
	
	RBTexture tex;
	tex.type = type;
	tex.format = format;
	tex.width = width;
	tex.height = height;
	tex.sRGB = false;
	tex.genMips = genMips;
	
	switch( format )
	{
	case TextureFormats::BGRA8:
#ifdef RENDERER_OPENGL15
		tex.glFmt = GL_RGBA8;
#else
		tex.glFmt = GL_RGBA;
#endif
	break;
	case TextureFormats::RGB5_A1:
#ifdef RENDERER_OPENGL15
		tex.glFmt = GL_RGB5_A1;
#else
		tex.glFmt = GL_RGBA;
#endif
	break;
	case TextureFormats::RGB565:
#ifdef RENDERER_OPENGL15
		tex.glFmt = GL_RGB565;
#else
		tex.glFmt = GL_RGB;
#endif
	break;
	case TextureFormats::RGBA4:
#ifdef RENDERER_OPENGL15
		tex.glFmt = GL_RGBA4;
#else
		tex.glFmt = GL_RGBA;
#endif
	break;
	case TextureFormats::DXT1:
		tex.glFmt = GL_COMPRESSED_RGBA_S3TC_DXT1_EXT;
		break;
	case TextureFormats::DXT3:
		tex.glFmt = GL_COMPRESSED_RGBA_S3TC_DXT3_EXT;
		break;
	case TextureFormats::DXT5:
		tex.glFmt = GL_COMPRESSED_RGBA_S3TC_DXT5_EXT;
		break;
	case TextureFormats::RGBA16F:
		Modules::log().writeWarning( "Unsupported texture format: RGBA16F" );
		return 0;
	case TextureFormats::RGBA32F:
		Modules::log().writeWarning( "Unsupported texture format: RGBA32F" );
		return 0;
	case TextureFormats::DEPTH:
		Modules::log().writeWarning( "Depth texture is unsupported" );
		return 0;
	case TextureFormats::PVRTC2:
		tex.glFmt = GL_COMPRESSED_RGB_PVRTC_2BPPV1_IMG;
		break;
	case TextureFormats::PVRTC4:
		tex.glFmt = GL_COMPRESSED_RGB_PVRTC_4BPPV1_IMG;
		break;
	case TextureFormats::PVRTCA2:
		tex.glFmt = GL_COMPRESSED_RGBA_PVRTC_2BPPV1_IMG;
		break;
	case TextureFormats::PVRTCA4:
		tex.glFmt = GL_COMPRESSED_RGBA_PVRTC_4BPPV1_IMG;
		break;
	default:
		ASSERT( 0 );
		break;
	};
	
	glGenTextures( 1, &tex.glObj );
	glActiveTexture( GL_TEXTURE0 );
	glBindTexture( tex.type, tex.glObj );
	
//	if( genMips )
//		glTexParameteri( tex.type, GL_GENERATE_MIPMAP, GL_TRUE );
	if( hasMips || genMips )
		glTexParameteri( tex.type, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR );
	else
		glTexParameteri( tex.type, GL_TEXTURE_MIN_FILTER, GL_LINEAR );

	bindTexture( 0, 0 );

	// Calculate memory requirements
	tex.memSize = calcTextureSize( format, width, height );
	if( hasMips || genMips ) tex.memSize += ftoi_r( tex.memSize * 1.0f / 3.0f );
	if( type == TextureTypes::TexCube ) tex.memSize *= 6;
	_textureMem += tex.memSize;

	return _textures.add( tex );
}


void RendererBase::uploadTextureData( uint32 texObj, int slice, int mipLevel, const void *pixels )
{
	if (texObj == 0) return;
	const RBTexture &tex = _textures.getRef( texObj );

	if (tex.type == TextureTypes::TexCube)
		return;

	TextureFormats::List format = tex.format;

	bindTexture( 15, texObj );
	
	int inputFormat = GL_BGRA, inputType = GL_UNSIGNED_BYTE;
	int target = GL_TEXTURE_2D;
	
	switch( format )
	{
    case TextureFormats::RGB5_A1:
        inputFormat = GL_RGBA;
        inputType = GL_UNSIGNED_SHORT_5_5_5_1;
    break;
    case TextureFormats::RGB565:
        inputFormat = GL_RGB;
        inputType = GL_UNSIGNED_SHORT_5_6_5;
    break;
    case TextureFormats::RGBA4:
        inputFormat = GL_RGBA;
        inputType = GL_UNSIGNED_SHORT_4_4_4_4;
    break;
	case TextureFormats::RGBA16F:
		Modules::log().writeWarning( "Unsupported texture format: RGBA16F" );
		return;
	case TextureFormats::RGBA32F:
		Modules::log().writeWarning( "Unsupported texture format: RGBA32F" );
		return;
	case TextureFormats::DEPTH:
		Modules::log().writeWarning( "Depth texture format is unsupported" );
		return;
	};
	
	// Calculate size of next mipmap using "floor" convention
	int width = std::max( tex.width >> mipLevel, 1 ), height = std::max( tex.height >> mipLevel, 1 );
	
	if( format == TextureFormats::DXT1 || 
		format == TextureFormats::DXT3 || 
		format == TextureFormats::DXT5 ||
        format == TextureFormats::PVRTC2 ||
        format == TextureFormats::PVRTC4 ||
        format == TextureFormats::PVRTCA2 ||
        format == TextureFormats::PVRTCA4 )
		glCompressedTexImage2D( target, mipLevel, tex.glFmt, width, height, 0,
		                        calcTextureSize( format, width, height ), pixels );	
	else
		glTexImage2D( target, mipLevel, tex.glFmt, width, height, 0, inputFormat, inputType, pixels );

	if( tex.genMips )
	{
		glGenerateMipmapOES( GL_TEXTURE_2D );
	}

}


void RendererBase::releaseTexture( uint32 texObj )
{
	if( texObj == 0 ) return;
	
	const RBTexture &tex = _textures.getRef( texObj );
	glDeleteTextures( 1, &tex.glObj );

	_textureMem -= tex.memSize;
	_textures.remove( texObj );
}


void RendererBase::updateTextureData( uint32 texObj, int slice, int mipLevel, const void *pixels )
{
	const RBTexture &tex = _textures.getRef( texObj );
	
	bindTexture( 0, texObj );
	glTexParameteri( tex.type, GL_GENERATE_MIPMAP, GL_FALSE );
	
	uploadTextureData( texObj, slice, mipLevel, pixels );

	bindTexture( 0, 0x0 );
}


bool RendererBase::getTextureData( uint32 texObj, int slice, int mipLevel, void *buffer )
{
	Modules::log().writeWarning( "getTextureData is not supported with OpenGL ES 1.1 renderer" );
	return false;
}


void RendererBase::bindTexture( uint32 unit, uint32 texObj )
{
	ASSERT( unit < 16 );
	
	glActiveTexture( GL_TEXTURE0 + unit );
	
	if( texObj == 0 )
	{
		glBindTexture( GL_TEXTURE_2D, 0 );
	}
	else
	{
		const RBTexture &tex = _textures.getRef( texObj );
		glBindTexture( tex.type, tex.glObj );
	}

	_texSlots[unit].texObj = texObj;
}


// =================================================================================================
// Shaders
// =================================================================================================

uint32 RendererBase::loadShader( const char *vertexShader, const char *fragmentShader )
{
	int infologLength = 0;
	int charsWritten = 0;
	char *infoLog = 0x0;
	int status;

	_shaderLog = "";

	uint32 program = glCreateProgram();

	// Compile and Link Vertex and fragment shaders
	status = glCompileShaders(program, vertexShader, fragmentShader);

	if( !status )
	{
		// Get info
		glGetProgramiv( program, GL_VERTEX_INFO_LOG_LENGTH, &infologLength );
		if( infologLength > 1 )
		{
			infoLog = new char[infologLength];
			glGetShaderInfoLog( program, GL_VERTEX_SHADER, infologLength, &charsWritten, infoLog );
			_shaderLog = _shaderLog + "[Vertex Shader]\n" + infoLog;
			delete[] infoLog; infoLog = 0x0;
		}

		glGetProgramiv( program, GL_FRAGMENT_INFO_LOG_LENGTH, &infologLength );
		if( infologLength > 1 )
		{
			infoLog = new char[infologLength];
			glGetShaderInfoLog( program, GL_FRAGMENT_SHADER, infologLength, &charsWritten, infoLog );
			_shaderLog = _shaderLog + "[Fragment Shader]\n" + infoLog;
			delete[] infoLog; infoLog = 0x0;
		}

		glDeleteProgram( program );
		return 0;
	}

	return program;
}



uint32 RendererBase::createShader( const char *vertexShader, const char *fragmentShader )
{
	// Compile and link shader
	uint32 shdObj = RendererBase::loadShader( vertexShader, fragmentShader );
	if( shdObj == 0 ) return 0;

	int attribCount;
	glGetProgramiv( shdObj, GL_ACTIVE_ATTRIBUTES, &attribCount );
	
	// Find vertex layouts that have the same input signature as the shader
	for( uint32 i = 0; i < _vertexLayouts._objects.size(); ++i )
	{
		RBVertexLayout &vl = _vertexLayouts._objects[i];

		RBVertexLayout::ShaderData shdData;
		shdData.elemAttribIndices.resize( vl.elems.size(), -1 );

		bool matching = true;
		for( int j = 0; j < attribCount; ++j )
		{
			uint32 type, attr;
			char name[32];
			glGetActiveAttrib( shdObj, j, &attr, &type, 32, 0x0, name );

			bool found = false;
			for( uint32 k = 0; k < vl.elems.size(); ++k )
			{
				if( vl.elems[k].semanticName == name )
				{	
					if (shdData.elemAttribIndices[k]==-1)
						shdData.elemAttribIndices[k] = 0;
					shdData.elemAttribIndices[k] |= attr;
					found = true;
					break;
				}
			}

			if( !found )
			{	
				matching = false;
				break;
			}
		}

		// Store shader attribute bindings
		if( matching ) vl.shaderData[shdObj] = shdData;
	}

	return shdObj;
}


void RendererBase::releaseShader( uint32 shdObj )
{
	if( shdObj == 0 ) return;
	
	// Remove shader data from vertex layouts
	for( uint32 i = 0; i < _vertexLayouts._objects.size(); ++i )
	{
		std::map< uint32, RBVertexLayout::ShaderData >::iterator itr =
			_vertexLayouts._objects[i].shaderData.find( shdObj );
		if( itr != _vertexLayouts._objects[i].shaderData.end() )
			_vertexLayouts._objects[i].shaderData.erase( itr );
	}

	glDeleteProgram( shdObj );
}


void RendererBase::bindShader( uint32 shdObj )
{
	_curShaderObj = shdObj;
	glUseProgram( shdObj );
} 


int RendererBase::getShaderVar( uint32 shdObj, const char *var )
{
	return glGetUniformLocation( shdObj, var );
}


bool RendererBase::setShaderVar1i( uint32 shdObj, const char *var, int value )
{
	Modules::log().writeWarning( "setShaderVar1i is not supported with OpenGL ES 1.1 renderer" );
	return false;
}


// =================================================================================================
// Renderbuffers
// =================================================================================================

uint32 RendererBase::createRenderBuffer( uint32 width, uint32 height, TextureFormats::List format,
                                         bool depth, uint32 numColBufs, uint32 samples )
{
	if( (format == TextureFormats::RGBA16F || format == TextureFormats::RGBA32F) &&
		!_caps[RenderCaps::Tex_Float] )
	{
		return 0;
	}

	if( numColBufs > RBRenderBuffer::MaxColorAttachmentCount ) return 0;

	if( samples > 0 )
	{
		Modules::log().writeWarning( "GPU does not support desired multisampling quality for render target" );
	}

	RBRenderBuffer rb;
	rb.width = width;
	rb.height = height;

	// Create framebuffers
	glGenFramebuffersOES( 1, &rb.fbo );

	if( numColBufs > 0 )
	{
		// Attach color buffers
		for( uint32 j = 0; j < numColBufs; ++j )
		{
			glBindFramebufferOES( GL_FRAMEBUFFER_OES, rb.fbo );
			// Create a color texture
			uint32 texObj = createTexture( TextureTypes::Tex2D, rb.width, rb.height, format, false, false, false, false );
			ASSERT( texObj != 0 );
			uploadTextureData( texObj, 0, 0, 0x0 );
			rb.colTexs[j] = texObj;
			RBTexture &tex = _textures.getRef( texObj );
			// Attach the texture
			glFramebufferTexture2DOES( GL_FRAMEBUFFER_OES, GL_COLOR_ATTACHMENT0_OES + j, GL_TEXTURE_2D, tex.glObj, 0 );
		}
	}
	else
	{
		glBindFramebufferOES( GL_FRAMEBUFFER_OES, rb.fbo );
//TODO: what to do instead of this?
//		glDrawBuffer( GL_NONE );
//		glReadBuffer( GL_NONE );
	}

	// Attach depth buffer
	if( depth )
	{
		glBindFramebufferOES( GL_FRAMEBUFFER_OES, rb.fbo );
		// Create a depth renderbuffer
		glGenRenderbuffersOES( 1, &rb.depthBuf );
		glBindRenderbufferOES( GL_RENDERBUFFER_OES, rb.depthBuf );
		glRenderbufferStorageOES( GL_RENDERBUFFER_OES, _depthFormat, rb.width, rb.height );
		// Attach the renderbuffer
		glFramebufferRenderbufferOES( GL_FRAMEBUFFER_OES, GL_DEPTH_ATTACHMENT_OES,
		                              GL_RENDERBUFFER_OES, rb.depthBuf );
	}

	uint32 rbObj = _rendBufs.add( rb );
	
	// Check if FBO is complete
	bool valid = true;
	glBindFramebufferOES( GL_FRAMEBUFFER_OES, rb.fbo );
	uint32 status = glCheckFramebufferStatusOES( GL_FRAMEBUFFER_OES );
	glBindFramebufferOES( GL_FRAMEBUFFER_OES, _defaultFramebuffer );
	if( status != GL_FRAMEBUFFER_COMPLETE_OES ) valid = false;
	
	if( !valid )
	{
		releaseRenderBuffer( rbObj );
		return 0;
	}
	
	return rbObj;
}


void RendererBase::releaseRenderBuffer( uint32 rbObj )
{
	if (rbObj==0) return;
	RBRenderBuffer &rb = _rendBufs.getRef( rbObj );
	
	glBindFramebufferOES( GL_FRAMEBUFFER_OES, _defaultFramebuffer );
	
	if( rb.depthBuf != 0 ) glDeleteRenderbuffersOES( 1, &rb.depthBuf );
	rb.depthBuf = 0;
		
	for( uint32 i = 0; i < RBRenderBuffer::MaxColorAttachmentCount; ++i )
	{
		if( rb.colTexs[i] != 0 ) releaseTexture( rb.colTexs[i] );
		rb.colTexs[i] = 0;
	}

	if( rb.fbo != 0 ) glDeleteFramebuffersOES( 1, &rb.fbo );
	rb.fbo = 0;

	_rendBufs.remove( rbObj );
}


uint32 RendererBase::getRenderBufferTex( uint32 rbObj, uint32 bufIndex )
{
	RBRenderBuffer &rb = _rendBufs.getRef( rbObj );
	
	if( bufIndex < RBRenderBuffer::MaxColorAttachmentCount ) return rb.colTexs[bufIndex];
	else return 0;
}


void RendererBase::resolveRenderBuffer( uint32 rbObj )
{
	return;	// used for multisampling which is not supported
}


void RendererBase::setRenderBuffer( uint32 rbObj )
{
	// Resolve render buffer if necessary
	if( _curRendBuf != 0 ) resolveRenderBuffer( _curRendBuf );
	
	// Set new render buffer
	_curRendBuf = rbObj;
	
	if( rbObj == 0 )
	{
		glBindFramebufferOES( GL_FRAMEBUFFER_OES, _defaultFramebuffer );
		glViewport( _vpX, _vpY, _vpWidth, _vpHeight );
		_fbWidth = _vpWidth + _vpX;
		_fbHeight = _vpHeight + _vpY;
	}
	else
	{
		// Unbind all textures to make sure that no FBO attachment is bound any more
		for( uint32 i = 0; i < 16; ++i ) bindTexture( i, 0 );
		
		RBRenderBuffer &rb = _rendBufs.getRef( rbObj );

		glBindFramebufferOES( GL_FRAMEBUFFER_OES, rb.fbo );
		ASSERT( glCheckFramebufferStatusOES( GL_FRAMEBUFFER_OES ) == GL_FRAMEBUFFER_COMPLETE_OES );
		glViewport( 0, 0, rb.width, rb.height );
		_fbWidth = rb.width;
		_fbHeight = rb.height;
	}
}


bool RendererBase::getRenderBufferData( uint32 rbObj, int bufIndex, int *width, int *height,
                                        int *compCount, float *dataBuffer, int bufferSize )
{
	int x, y, w, h;

	if ((unsigned)bufIndex > RBRenderBuffer::MaxColorAttachmentCount-1)
		return false;
	
	if( rbObj == 0 )
	{
		if( bufIndex != 0 ) return false;
		if( width != 0x0 ) *width = _vpWidth;
		if( height != 0x0 ) *height = _vpHeight;
		
		x = _vpX; y = _vpY; w = _vpWidth; h = _vpHeight;

		glBindFramebufferOES( GL_FRAMEBUFFER_OES, _defaultFramebuffer );
	}
	else
	{
		resolveRenderBuffer( rbObj );
		RBRenderBuffer &rb = _rendBufs.getRef( rbObj );
		
		if( (unsigned)bufIndex > RBRenderBuffer::MaxColorAttachmentCount-1 || rb.colTexs[bufIndex] == 0 ) return false;
		if( width != 0x0 ) *width = rb.width;
		if( height != 0x0 ) *height = rb.height;

		x = 0; y = 0; w = rb.width; h = rb.height;
		
		glBindFramebufferOES( GL_FRAMEBUFFER_OES, rb.fbo );
	}
	
	int comps = 1; // (bufIndex == 32 ? 1 : 4);
	if( compCount != 0x0 ) *compCount = comps;
	
	if( dataBuffer == 0x0 ) return true;
	if( bufferSize < w * h * comps * (signed)sizeof( float ) ) return false;
	
	glFinish();
	glReadPixels( x, y, w, h, GL_RGBA, GL_UNSIGNED_BYTE, dataBuffer );
	
	return true;
}


// =================================================================================================
// Queries
// =================================================================================================

uint32 RendererBase::createOcclusionQuery()
{
	return 0;
}


void RendererBase::releaseQuery( uint32 /*queryObj*/ )
{
}


void RendererBase::beginQuery( uint32 /*queryObj*/ )
{
}


void RendererBase::endQuery( uint32 /*queryObj*/ )
{
}


uint32 RendererBase::getQueryResult( uint32 /*queryObj*/ )
{
	return 0;
}


// =================================================================================================
// Vertex declarations
// =================================================================================================

uint32 RendererBase::createVertexLayout( uint32 elemCount )
{
	RBVertexLayout vl;
	
	uint32 vlObj = _vertexLayouts.add( vl );
	_vertexLayouts.getRef( vlObj ).elems.resize( elemCount );

	return vlObj;
}

void RendererBase::setVertexLayoutElem( uint32 vlObj, uint32 slot, const char *semanticName,
                                        uint32 vbSlot, uint32 size, uint32 offset )
{
	RBVertexLayout &vl = _vertexLayouts.getRef( vlObj );
	
	vl.elems[slot].semanticName = semanticName;
	vl.elems[slot].vbSlot = vbSlot;
	vl.elems[slot].size = size;
	vl.elems[slot].offset = offset;
}

void RendererBase::releaseVertexLayout( uint32 vlObj )
{
	if( vlObj == 0 ) return;
	
	_vertexLayouts.remove( vlObj );
}

bool RendererBase::applyVertexLayout( uint32 vlObj )
{
	// Notes:
	// Vertex layouts contain shader-specific data.
	// Vertex layouts need to be applied after the desired shader has been bound and all required
	// vertex buffers were bound!
	
	// Reset previous layout
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);
	for(int i=0;i<2;++i)	//at least two is always supported (using only two for now)
	{
		glClientActiveTexture(GL_TEXTURE0 + i);
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	}
	glClientActiveTexture(GL_TEXTURE0);
	glBindBuffer( GL_ARRAY_BUFFER, 0);
	
	if( vlObj == 0 ) return false;

	RBVertexLayout &vl = _vertexLayouts.getRef( vlObj );

	// Find data for the currently bound shader
	ASSERT( _curShaderObj != 0 );
	std::map< uint32, RBVertexLayout::ShaderData >::iterator itr = vl.shaderData.find( _curShaderObj );
	if( itr == vl.shaderData.end() ) return false;

	// Set vertex attrib pointers
	for( uint32 i = 0; i < vl.elems.size(); ++i )
	{
		RBVertLayoutElem &elem = vl.elems[i];
		const RBVertBufSlot &vbSlot = _vertBufSlots[elem.vbSlot];
		
		int attribIndex = itr->second.elemAttribIndices[i];
		if( attribIndex >= 0 )
		{
			ASSERT( _buffers.getRef( _vertBufSlots[elem.vbSlot].vbObj ).glObj != 0 &&
					_buffers.getRef( _vertBufSlots[elem.vbSlot].vbObj ).type == GL_ARRAY_BUFFER );
			glBindBuffer( GL_ARRAY_BUFFER, _buffers.getRef( _vertBufSlots[elem.vbSlot].vbObj ).glObj );

			if (attribIndex&GL_ATTRIB_VERTEX)	//vertex
			{
				glVertexPointer( elem.size, GL_FLOAT,
								   vbSlot.stride, (char *)0 + vbSlot.offset + elem.offset );
				glEnableClientState( GL_VERTEX_ARRAY );
			}
			if (attribIndex&GL_ATTRIB_NORMAL)	//color
			{
				glNormalPointer( GL_FLOAT,
								   vbSlot.stride, (char *)0 + vbSlot.offset + elem.offset );
				glEnableClientState( GL_NORMAL_ARRAY );
			}
			if (attribIndex&GL_ATTRIB_COLOR)	//normal
			{
				glColorPointer( elem.size, GL_FLOAT,
								   vbSlot.stride, (char *)0 + vbSlot.offset + elem.offset );
				glEnableClientState( GL_COLOR_ARRAY );
			}
			if (attribIndex&GL_ATTRIB_MULTITEXCOORD0)	//texcoord0
			{
				glTexCoordPointer( elem.size, GL_FLOAT,
								   vbSlot.stride, (char *)0 + vbSlot.offset + elem.offset );
				glEnableClientState( GL_TEXTURE_COORD_ARRAY );
			}
			if (attribIndex&GL_ATTRIB_MULTITEXCOORD1)	//texcoord1
			{
				glClientActiveTexture(GL_TEXTURE1);
				glTexCoordPointer( elem.size, GL_FLOAT,
								   vbSlot.stride, (char *)0 + vbSlot.offset + elem.offset );
				glEnableClientState( GL_TEXTURE_COORD_ARRAY );
				glClientActiveTexture(GL_TEXTURE0);
			}
		}
	}

	return true;
}
