#include "oglExtensions.h"

#include "utOpenGLES11.h"

#include <stdlib.h>
#include <string.h>
#include <string>
#include <algorithm>
#include <math.h>
#include <vector>
#include <assert.h>

//
//	TODO: support #define and #ifdef
//

typedef unsigned int uint32;

	//
	//	helper classes from horde3d
	//

class cTokenizer
{
protected:	
	
	static const ptrdiff_t tokenSize = 128;
	
	char        _token[tokenSize], _prevToken[tokenSize];
	const char  *_p;
	int         _line;

	const char*	_whitespaceandspecial;


	void checkLineChange()
	{
		if( *_p == '\r' && *(_p+1) == '\n' )
		{
			++_p;
			++_line;
		}
		else if( *_p == '\r' || *_p == '\n' ) ++_line;
	}

	void skip( const char *chars )
	{
		while( *_p )
		{
			if( !strchr( chars, *_p ) ) break;
			checkLineChange();
			++_p;
		}
	}
	
	bool seekChar( const char *chars )
	{
		while( *_p )
		{
			if( strchr( chars, *_p ) ) break;
			checkLineChange();
			++_p;
		}
		return *_p != '\0';
	}
	
	void getNextToken()
	{
		// Skip whitespace
		skip( " \t\n\r" );

		// Parse token
		const char *p0 = _p, *p1 = _p;
		if( *_p == '"' )  // Handle string
		{
			++_p; ++p0;
			if( seekChar( "\"\n\r" ) ) p1 = _p++;
		}
		else
		{
			seekChar( _whitespaceandspecial );  // Advance until whitespace or special char found
			if( _p == p0 && *_p != '\0' ) ++_p;  // Handle special char
			p1 = _p;
		}
		memcpy( _token, p0, std::min( (ptrdiff_t)(p1 - p0), tokenSize-1 ) );
		_token[std::min( (ptrdiff_t)(p1 - p0), tokenSize-1 )] = '\0';
	}

public:
	
	cTokenizer( const char *data ) : _p( data ), _line( 1 ), _whitespaceandspecial(" \t\n\r{}()[]<>=,;") { getNextToken(); }

	int getLine() { return _line; }

	bool hasToken() { return _token[0] != '\0'; }

	bool checkToken( const char *token, bool peekOnly = false )
	{
		if( _stricmp( _token, token ) == 0 )
		{
			if( !peekOnly ) getNextToken();
			return true;
		}
		return false;
	}

	const char *getToken( const char *charset )
	{
		if( charset )
		{
			// Validate token
			const char *p = _token;
			while( *p )
			{
				if( strchr( charset, *p++ ) == 0x0 )
				{
					_prevToken[0] = '\0';
					return _prevToken;
				}
			}
		}
		
		memcpy( _prevToken, _token, tokenSize );
		getNextToken();
		return _prevToken;
	}

	bool seekToken( const char *token )
	{
		while( _stricmp( getToken( 0x0 ), token ) != 0 )
		{
			if( !hasToken() ) return false;
		}
		return true;
	}

	const char* getWhitespaceAndSpecial() { return _whitespaceandspecial; }
	void setWhitespaceAndSpecial(const char* s) { _whitespaceandspecial = s; }
};

template< class T > class RBObjects
{
private:
	std::vector< T >       _objects;
	std::vector< uint32 >  _freeList;

public:
	uint32 add( T &obj )
	{
		if( !_freeList.empty() )
		{
			uint32 index = _freeList.back();
			_freeList.pop_back();
			_objects[index] = obj;
			return index + 1;
		}
		else
		{
			_objects.push_back( obj );
			return (uint32)_objects.size();
		}
	}

	void remove( uint32 handle )
	{
		ASSERT( handle > 0 && handle <= _objects.size() );
		
		_objects[handle - 1] = T();  // Destruct and replace with default object
		_freeList.push_back( handle - 1 );
	}

	T* getRef( uint32 handle )
	{
		if ( handle > _objects.size() )
			return NULL;
		
		return &_objects[handle - 1];
	}

	friend class RendererBase;
};

	//upper 8 bit is for index
struct eFFStates
{
	enum Enum
	{
			//vertex data
		Color,	// vec4
		ColorMaterialEnabled, // True|False;
		NormalizeEnabled, // True|False;
		RescaleNormalEnabled, // True|False;

			//Fog paremeters
		FogEnabled, // True|False;	(default = false)
		FogMode, // = Linear|Exp|Exp2;
		FogDensity, // float
		FogStart, // float
		FogEnd, // float
		FogColor, // vec4

			//Lighting
		LightingEnabled, // = True|False; (default = false)
		LightModelAmbient, // vec4
		LightModelTwoSide, // True|False;

			// Light parameters for MAX_LIGHTS number of lights
		LightEnabled,	//[MAX_LIGHTS] = True|False; (default = false only if Lighting is enabled)
		LightAmbient, //[MAX_LIGHTS] vec4
		LightDiffuse,	//[MAX_LIGHTS] vec4
		LightSpecular, //[MAX_LIGHTS] vec4
		LightEmission, //[MAX_LIGHTS] vec4
		LightPosition, //[MAX_LIGHTS] vec4
		LightSpotDirection, //[MAX_LIGHTS] vec3
		LightSpotExponent, //[MAX_LIGHTS] float
		LightSpotCutoff, //[MAX_LIGHTS] float
		LightConstantAttenuation, //[MAX_LIGHTS] float
		LightLinearAttenuation, //[MAX_LIGHTS] float
		LightQuadraticAttenuation, //[MAX_LIGHTS] float

			//Material
		MaterialAmbient, //vec4;
		MaterialDiffuse, // vec4
		MaterialSpecular, // vec4
		MaterialEmission, //  vec4
		MaterialShininess, // float

			//transform
		ModelViewMatrix,	// mat4
		ProjectionMatrix,	// mat4
		TextureMatrix,	//[MAX_TEXTURE_UNITS] mat4

		// Fragment pipeline
			// alpha test
		AlphaTestEnabled, //True|False (default = false)
		AlphaFunc, // Never|Less|Equal|LEqual|Greater|NotEqual|GEqual|Always
		AlphaRef,  // float

			//states for each MAX_TEXTURE_UNITS
		Texture2DEnabled,	// (default = false) -> if true sets texture envmode to combine
		Constant, // [n] = vec4;
		CombineRgb, // Replace|Modulate|Add|AddSigned|Interpolate|Subtract|Dot3
		CombineAlpha, //Replace|Modulate|Add|AddSigned|Interpolate|Subtract 
		Src0Rgb, 
		Src1Rgb,
		Src2Rgb,
		Src0Alpha,
		Src1Alpha,
		Src2Alpha,
		Operand0Rgb,
		Operand1Rgb,
		Operand2Rgb,
		Operand0Alpha,
		Operand1Alpha,
		Operand2Alpha,
		RgbScale,	//float: 1,2,4
		AlphaScale, //float: 1,2,4
	};
};

#define MAKE_INDEXED_STATE(state,index)	(state|(index<<24))
#define GET_STATE(indexedstate) (indexedstate&0x00ffffff)
#define GET_INDEX(indexedstate) ((indexedstate>>24)&0xff)

struct sFloatConstant
{
	sFloatConstant(int state, float v) : mState(state), mValue(v)
	{}
	int		mState;
	float	mValue;
};
struct sVec3Constant
{
	sVec3Constant(int state, const float* v)
	{
		mState = state;
		for(unsigned int i=0; i<3; ++i)
			mValue[i] = v[i];
	}

	int		mState;
	float	mValue[3];
};
struct sVec4Constant
{
	sVec4Constant(int state, const float* v)
	{
		mState = state;
		for(unsigned int i=0; i<4; ++i)
			mValue[i] = v[i];
	}
	int		mState;
	float	mValue[4];
};

struct sMat4Constant
{
	sMat4Constant(int state, const float* v)
	{
		mState = state;
		for(unsigned int i=0; i<16; ++i)
			mValue[i] = v[i];
	}
	int		mState;
	float	mValue[16];
};
struct sIntConstant
{
	sIntConstant(int state, int v) : mState(state), mValue(v)
	{}
	int		mState;
	int		mValue;
};

struct sUniform
{
	sUniform(const char* name, int type): mName(name), mType(type)
	{}
	int					mType;//float,float3,...
	std::vector<int>	mStates;
	std::string			mName;
};

struct sAttribute
{
	sAttribute(int id, int type, const char* semantic): mID(id), mType(type), mSemantic(semantic)
	{}
	int				mID;	//GL_ATTRIB_VERTEX,GL_ATTRIB_NORMAL,...
	int				mType;	//GL_FLOAT2,...
	std::string		mSemantic;	//special info for client (eg what to put into this attribute)
};

struct FixedFunctionProgram
{
	int	raiseError(int mode, const char *msg, ... )
	{
		va_list args;
		va_start( args, msg );
		char _textBuf[2048];
		vsnprintf( _textBuf, 2048, msg, args );
		va_end( args );

		if (mode == GL_VERTEX_SHADER)
			mVertexInfo += _textBuf;
		else if (mode == GL_FRAGMENT_SHADER)
			mFragmentInfo += _textBuf;

		return 0;
	}
	int	raiseFSError(const char* message, int line);

		//constant states
	std::vector<sFloatConstant>	mFloatConstants;
	std::vector<sVec3Constant>	mVec3Constants;
	std::vector<sVec4Constant>	mVec4Constants;
	std::vector<sMat4Constant>	mMat4Constants;
	std::vector<sIntConstant>	mIntConstants;

		//uniform -> state mappings
	std::vector<sUniform>		mUniforms;

	//attributes
	std::vector<sAttribute>		mAttributes;

	std::string					mVertexInfo;
	std::string					mFragmentInfo;

	// functions for TextureMatrix?
};

static RBObjects< FixedFunctionProgram >  gPrograms;
static int gCurrentProgram = 0;
GLint glCompileVertexShader(FixedFunctionProgram* prg, char* source);
GLint glCompileFragmentShader(FixedFunctionProgram* prg, char* source);
static bool handleFloatAssignment(FixedFunctionProgram* prg, cTokenizer& tok, int state);
static bool handleVec3Assignment(FixedFunctionProgram* prg, cTokenizer& tok, int state);
static bool handleVec4Assignment(FixedFunctionProgram* prg, cTokenizer& tok, int state);
static bool handleMat4Assignment(FixedFunctionProgram* prg, cTokenizer& tok, int state);
static int handleReadIndex(cTokenizer& tok);
static void removeComments(char* text);
static void preprocessMacro(char *text);

GLuint glCreateProgram(void)
{
	FixedFunctionProgram ffs;
	GLuint program = gPrograms.add(ffs);
	return program;
}

void glDeleteProgram (GLuint program)
{
	gPrograms.remove(program);
}

void glUseProgram (GLuint program)
{
	if (program == 0)
	{	//set default state
		glDisable(GL_FOG);
		glDisable(GL_LIGHTING);
		glDisable(GL_NORMALIZE);
		glDisable(GL_RESCALE_NORMAL);
		glDisable(GL_ALPHA_TEST);

		glActiveTexture(GL_TEXTURE1);
		glDisable(GL_TEXTURE_2D);

		glActiveTexture(GL_TEXTURE0);
		glDisable(GL_TEXTURE_2D);
		glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
		return;	//TODO: disable shaders?
	}

	gCurrentProgram = program;

	FixedFunctionProgram* p = gPrograms.getRef(program);

	// setup constant states
	for(unsigned int i=0; i<p->mFloatConstants.size(); ++i)
	{
		sFloatConstant& c = p->mFloatConstants[i];

		switch(GET_STATE(c.mState))
		{
		case eFFStates::FogDensity:
			glFogf(GL_FOG_DENSITY, c.mValue);
		break;
		case eFFStates::FogStart:
			glFogf(GL_FOG_START, c.mValue);
		break;
		case eFFStates::FogEnd:
			glFogf(GL_FOG_END, c.mValue);
		break;
		case eFFStates::FogMode:
			glFogf(GL_FOG_MODE, c.mValue);
		break;

		case eFFStates::LightModelTwoSide:
			glLightModelf(GL_LIGHT_MODEL_TWO_SIDE, c.mValue);
		break;

		case eFFStates::LightSpotExponent:
			glLightf(GL_LIGHT0 + GET_INDEX(c.mState), GL_SPOT_EXPONENT, c.mValue);
		break;
		case eFFStates::LightSpotCutoff:
			glLightf(GL_LIGHT0 + GET_INDEX(c.mState), GL_SPOT_CUTOFF, c.mValue);
		break;
		case eFFStates::LightConstantAttenuation:
			glLightf(GL_LIGHT0 + GET_INDEX(c.mState), GL_CONSTANT_ATTENUATION, c.mValue);
		break;
		case eFFStates::LightLinearAttenuation:
			glLightf(GL_LIGHT0 + GET_INDEX(c.mState), GL_LINEAR_ATTENUATION, c.mValue);
		break;
		case eFFStates::LightQuadraticAttenuation:
			glLightf(GL_LIGHT0 + GET_INDEX(c.mState), GL_QUADRATIC_ATTENUATION, c.mValue);
		break;
		case eFFStates::MaterialShininess:
			glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, c.mValue);
		break;
		case eFFStates::AlphaRef:
		{
			int func;
			glGetIntegerv(GL_ALPHA_TEST_FUNC, &func);
			glAlphaFunc(func, c.mValue);
		}
		break;
		case eFFStates::RgbScale:
			glActiveTexture(GL_TEXTURE0 + GET_INDEX(c.mState));
			glTexEnvf(GL_TEXTURE_ENV, GL_RGB_SCALE, c.mValue);
		break;
		case eFFStates::AlphaScale:
			glActiveTexture(GL_TEXTURE0 + GET_INDEX(c.mState));
			glTexEnvf(GL_TEXTURE_ENV, GL_ALPHA_SCALE, c.mValue);
		break;
		default:
			;
			//invalid state
		}
	}

	for(unsigned int i=0; i<p->mVec3Constants.size(); ++i)
	{
		sVec3Constant& c = p->mVec3Constants[i];

		switch(GET_STATE(c.mState))
		{
		case eFFStates::LightSpotDirection:
			glLightfv(GL_LIGHT0 + GET_INDEX(c.mState), GL_SPOT_DIRECTION, c.mValue);
		break;
		default:
			;
			//invalid state
		}
	}

	for(unsigned int i=0; i<p->mVec4Constants.size(); ++i)
	{
		sVec4Constant& c = p->mVec4Constants[i];

		switch(GET_STATE(c.mState))
		{
		case eFFStates::Color:	// vec4
			glColor4f(c.mValue[0],c.mValue[1],c.mValue[2],c.mValue[3]);
		break;
		case eFFStates::FogColor: // vec4
			glFogfv(GL_FOG_COLOR, c.mValue);
		break;
		case eFFStates::LightModelAmbient: // vec4
			glLightModelfv(GL_LIGHT_MODEL_AMBIENT , c.mValue);
		break;
		case eFFStates::LightAmbient:
			glLightfv(GL_LIGHT0 + GET_INDEX(c.mState), GL_AMBIENT, c.mValue);
		break;
		case eFFStates::LightDiffuse:
			glLightfv(GL_LIGHT0 + GET_INDEX(c.mState), GL_DIFFUSE, c.mValue);
		break;
		case eFFStates::LightSpecular:
			glLightfv(GL_LIGHT0 + GET_INDEX(c.mState), GL_SPECULAR, c.mValue);
		break;
		case eFFStates::LightEmission:
			glLightfv(GL_LIGHT0 + GET_INDEX(c.mState), GL_EMISSION, c.mValue);
		break;
		case eFFStates::LightPosition:
			glMatrixMode(GL_MODELVIEW);
			glPushMatrix();
			glLoadIdentity();
			glLightfv(GL_LIGHT0 + GET_INDEX(c.mState), GL_POSITION, c.mValue);
			glPopMatrix();
		break;
		case eFFStates::MaterialAmbient:
			glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, c.mValue);
		break;
		case eFFStates::MaterialDiffuse:
			glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, c.mValue);
		break;
		case eFFStates::MaterialSpecular: // vec4
			glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, c.mValue);
		break;
		case eFFStates::MaterialEmission: //  vec4
			glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, c.mValue);
		break;

		// Fragment pipeline
		case eFFStates::Constant: // [n] = vec4Variable;
		{
			glActiveTexture(GL_TEXTURE0 + GET_INDEX(c.mState));
			glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, c.mValue);
		}
		break;
		}
	}

	for(unsigned int i=0; i<p->mMat4Constants.size(); ++i)
	{
		sMat4Constant& c = p->mMat4Constants[i];

		switch(GET_STATE(c.mState))
		{
		case eFFStates::ModelViewMatrix:
			glMatrixMode(GL_MODELVIEW);
			glLoadMatrixf(c.mValue);
		break;
		case eFFStates::ProjectionMatrix:
			glMatrixMode(GL_PROJECTION);
			glLoadMatrixf(c.mValue);
		break;
		case eFFStates::TextureMatrix:	//active texture unit??
			glMatrixMode(GL_TEXTURE);
			glLoadMatrixf(c.mValue);
		break;
		}
	}

	for(unsigned int i=0; i<p->mIntConstants.size(); ++i)
	{
		sIntConstant& c = p->mIntConstants[i];

		switch(GET_STATE(c.mState))
		{
		case eFFStates::ColorMaterialEnabled:
			if (c.mValue) glEnable(GL_COLOR_MATERIAL);
			else glDisable(GL_COLOR_MATERIAL);
		break;
		case eFFStates::NormalizeEnabled:
			if (c.mValue) glEnable(GL_NORMALIZE);
			else glDisable(GL_NORMALIZE);
		break;
		case eFFStates::RescaleNormalEnabled:
			if (c.mValue) glEnable(GL_RESCALE_NORMAL);
			else glDisable(GL_RESCALE_NORMAL);
		break;
		case eFFStates::FogEnabled:
			if (c.mValue) glEnable(GL_FOG);
			else glDisable(GL_FOG);
		break;
		case eFFStates::LightingEnabled:
			if (c.mValue) glEnable(GL_LIGHTING);
			else glDisable(GL_LIGHTING);
		break;

		case eFFStates::LightEnabled:	//[MAX_LIGHTS] = True|Fa
			if (c.mValue) glEnable(GL_LIGHT0 + GET_INDEX(c.mState));
			else glDisable(GL_LIGHT0 + GET_INDEX(c.mState));
		break;

		case eFFStates::AlphaTestEnabled: //True|False
			if (c.mValue) glEnable(GL_ALPHA_TEST);
			else glDisable(GL_ALPHA_TEST);
		break;
		case eFFStates::AlphaFunc: // Never|Less|Equal|LEqual|Greater|NotEqual|GEqual|Always
		{
			float ref;
			glGetFloatv(GL_ALPHA_TEST_FUNC, &ref);
			glAlphaFunc(c.mValue, ref);
		}
		break;

			//states for each MAX_TEXTURE_UNITS
		case eFFStates::Texture2DEnabled:
			glActiveTexture(GL_TEXTURE0 + GET_INDEX(c.mState));
			if (c.mValue) 
			{
				glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
				glEnable(GL_TEXTURE_2D);
			}
			else glDisable(GL_TEXTURE_2D);
		break;
		case eFFStates::CombineRgb:
			glActiveTexture(GL_TEXTURE0 + GET_INDEX(c.mState));
			glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB, c.mValue);
		break;
		case eFFStates::CombineAlpha: //Replace|Modulate|Add|AddSigned|Interpolate|Subtract 
			glActiveTexture(GL_TEXTURE0 + GET_INDEX(c.mState));
			glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_ALPHA, c.mValue);
		break;
		case eFFStates::Src0Rgb:
			glActiveTexture(GL_TEXTURE0 + GET_INDEX(c.mState));
			glTexEnvi(GL_TEXTURE_ENV, GL_SRC0_RGB, c.mValue);
		break;
		case eFFStates::Src1Rgb:
			glActiveTexture(GL_TEXTURE0 + GET_INDEX(c.mState));
			glTexEnvi(GL_TEXTURE_ENV, GL_SRC1_RGB, c.mValue);
		break;
		case eFFStates::Src2Rgb:
			glActiveTexture(GL_TEXTURE0 + GET_INDEX(c.mState));
			glTexEnvi(GL_TEXTURE_ENV, GL_SRC2_RGB, c.mValue);
		break;
		case eFFStates::Src0Alpha:
			glActiveTexture(GL_TEXTURE0 + GET_INDEX(c.mState));
			glTexEnvi(GL_TEXTURE_ENV, GL_SRC0_ALPHA, c.mValue);
		break;
		case eFFStates::Src1Alpha:
			glActiveTexture(GL_TEXTURE0 + GET_INDEX(c.mState));
			glTexEnvi(GL_TEXTURE_ENV, GL_SRC1_ALPHA, c.mValue);
		break;
		case eFFStates::Src2Alpha:
			glActiveTexture(GL_TEXTURE0 + GET_INDEX(c.mState));
			glTexEnvi(GL_TEXTURE_ENV, GL_SRC2_ALPHA, c.mValue);
		break;
		case eFFStates::Operand0Rgb:
			glActiveTexture(GL_TEXTURE0 + GET_INDEX(c.mState));
			glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB, c.mValue);
		break;
		case eFFStates::Operand1Rgb:
			glActiveTexture(GL_TEXTURE0 + GET_INDEX(c.mState));
			glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_RGB, c.mValue);
		break;
		case eFFStates::Operand2Rgb:
			glActiveTexture(GL_TEXTURE0 + GET_INDEX(c.mState));
			glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND2_RGB, c.mValue);
		break;
		case eFFStates::Operand0Alpha:
			glActiveTexture(GL_TEXTURE0 + GET_INDEX(c.mState));
			glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_ALPHA, c.mValue);
		break;
		case eFFStates::Operand1Alpha:
			glActiveTexture(GL_TEXTURE0 + GET_INDEX(c.mState));
			glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_ALPHA, c.mValue);
		break;
		case eFFStates::Operand2Alpha:
			glActiveTexture(GL_TEXTURE0 + GET_INDEX(c.mState));
			glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND2_ALPHA, c.mValue);
		break;
		}
	}
}

GLint glCompileShaders(GLuint program, const char* vertexShaderSource, const char* fragmentShaderSource)
{
	FixedFunctionProgram* prg = gPrograms.getRef(program);
	if (prg==NULL)
		return GL_FALSE;

	int retval = GL_TRUE;

	//parse and compile vertex/fragment shaders
	for(int i=0;i<2;++i)
	{
		const char* shaderSource = (i == 0) ? vertexShaderSource : fragmentShaderSource;

		char* data = strdup(shaderSource);

		// Preprocessing: Replace comments with whitespace
		removeComments(data);

		// Preprocessing: Replace # lines and comments with whitespace
		preprocessMacro(data);


		if ( i == 0)
		{
			if ( glCompileVertexShader(prg, data) == GL_FALSE)
				retval = GL_FALSE;
		}
		else
		{
			if ( glCompileFragmentShader(prg, data) == GL_FALSE)
				retval = GL_FALSE;
		}

		free(data);

		if (retval == GL_FALSE)
			return GL_FALSE;
	}

	// add default state constant settings (if not set explicitly)
	int states[5] = {eFFStates::FogEnabled, eFFStates::LightingEnabled,
					eFFStates::NormalizeEnabled,eFFStates::RescaleNormalEnabled,eFFStates::AlphaTestEnabled};
	int stateIndex[5] = {-1,-1,-1,-1,-1};
	for(int s = 0; s<5;++s)
	{
		for(unsigned int i=0; i< prg->mIntConstants.size();++i)
		{
			if (prg->mIntConstants[i].mState == states[s])
			{
				stateIndex[s] = i;
				break;
			}
		}
		if (stateIndex[s] == -1)
			prg->mIntConstants.push_back(sIntConstant(states[s], 0));
	}

	//disable each light in lighting is off and they are not used
	if (stateIndex[1]!=-1 && prg->mIntConstants[stateIndex[1]].mValue!=0)
	{	//disable each light that is not enabled
		for(int s = 0; s<8;++s)
		{
			bool found = false;
			for(unsigned int i=0; i< prg->mIntConstants.size();++i)
			{
				if (prg->mIntConstants[i].mState == MAKE_INDEXED_STATE(eFFStates::LightEnabled,s))
				{
					found = true;
					break;
				}
			}
			if (!found)
				prg->mIntConstants.push_back(sIntConstant(MAKE_INDEXED_STATE(eFFStates::LightEnabled,s), 0));
		}
	}
	
	//check the states disable the state after the last sequentially enabled
	int maxTextureUnit;
	glGetIntegerv(GL_MAX_TEXTURE_UNITS, &maxTextureUnit);
	//disable each light that is not enabled
	for(int s = 0; s<maxTextureUnit;++s)
	{
		int idx = -1;
		for(unsigned int i=0; i< prg->mIntConstants.size();++i)
		{
			if (prg->mIntConstants[i].mState == MAKE_INDEXED_STATE(eFFStates::Texture2DEnabled,s))
			{
				idx = i;
				break;
			}
		}

		if (idx == -1)
		{	//not found disable this
			prg->mIntConstants.push_back(sIntConstant(MAKE_INDEXED_STATE(eFFStates::Texture2DEnabled,s), 0));
			break;
		}
		else if (prg->mIntConstants[idx].mState == 0)
		{ //disabled no more work after this
			break;
		}
	}

	return retval;
}


GLint glCompileVertexShader(FixedFunctionProgram* prg, char* source)
{
	cTokenizer tok( source );
	const char *identifier = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_";
	const char *intnum = "+-0123456789";
	const char *floatnum = "+-0123456789.eE";

	while( tok.hasToken() )
	{
		if( tok.checkToken( "uniform" ) )
		{
			int uniformType = 0;
			if (tok.checkToken( "float") ) uniformType = GL_FLOAT;
			else if (tok.checkToken( "vec3") ) uniformType = GL_FLOAT_VEC3;
			else if (tok.checkToken( "vec4") ) uniformType = GL_FLOAT_VEC4;
			else if (tok.checkToken( "mat4") ) uniformType = GL_FLOAT_MAT4;
			else 
				return prg->raiseError(GL_VERTEX_SHADER, "Invalid uniform type: %s in line %d", tok.getToken(0), tok.getLine());

			const char* id = tok.getToken( identifier );
			
			prg->mUniforms.push_back(sUniform(id,uniformType));

		}
		else if( tok.checkToken( "attribute" ) )
		{
			int attributeType = 0;
			if (tok.checkToken( "float") ) attributeType = GL_FLOAT;
			else if (tok.checkToken( "vec2") ) attributeType = GL_FLOAT_VEC2;
			else if (tok.checkToken( "vec3") ) attributeType = GL_FLOAT_VEC3;
			else if (tok.checkToken( "vec4") ) attributeType = GL_FLOAT_VEC4;
			else  
				return prg->raiseError(GL_VERTEX_SHADER, "Invalid attribute type: %s in line %d", tok.getToken(0), tok.getLine());
			
				//attributeName
			int attributeName = 0;
			if (tok.checkToken( "gl_Vertex") ) attributeName = GL_ATTRIB_VERTEX;
			else if (tok.checkToken( "gl_Normal") ) attributeName = GL_ATTRIB_NORMAL;
			else if (tok.checkToken( "gl_Color") ) attributeName = GL_ATTRIB_COLOR;
			else if (tok.checkToken( "gl_MultiTexCoord0") ) attributeName = GL_ATTRIB_MULTITEXCOORD0;
			else if (tok.checkToken( "gl_MultiTexCoord1") ) attributeName = GL_ATTRIB_MULTITEXCOORD1;
			else 
				return prg->raiseError(GL_VERTEX_SHADER, "Invalid attribute name: %s in line %d", tok.getToken(0), tok.getLine());

			const char* annotation = "";

			if (tok.checkToken("=")) 
			{	//special annotation
				annotation = tok.getToken( identifier );
			}

			prg->mAttributes.push_back(sAttribute(attributeName, attributeType, annotation));
		}
		else if( tok.checkToken( "Color" ) )
		{
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());
			if (!handleVec4Assignment(prg, tok, eFFStates::Color))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected vec4 variable or literal in line %d", tok.getLine());
		}
		else if( tok.checkToken( "FogEnable" ) )
		{
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());

			int v = 0;
			if (tok.checkToken( "True")) v = 1;
			else if (tok.checkToken("False")) v = 0;
			else return prg->raiseError(GL_VERTEX_SHADER, "Expected True or False in line %d", tok.getLine());

			prg->mIntConstants.push_back(sIntConstant(eFFStates::FogEnabled, v));
		}
		else if( tok.checkToken( "FogMode" ) )
		{
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());

			int v = 0;
			if (tok.checkToken( "Linear")) v = GL_LINEAR;
			else if (tok.checkToken("Exp")) v = GL_EXP;
			else if (tok.checkToken("Exp2")) v = GL_EXP2;
			else return prg->raiseError(GL_VERTEX_SHADER, "Expected Linear|Exp|Exp2 in line %d", tok.getLine());

			prg->mIntConstants.push_back(sIntConstant(eFFStates::FogMode, v));
		}
		else if( tok.checkToken( "FogDensity" ) )
		{
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());
			if (!handleFloatAssignment(prg, tok, eFFStates::FogDensity))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected float variable or literal in line %d", tok.getLine());
		}
		else if( tok.checkToken( "FogStart" ) )
		{
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());
			if (!handleFloatAssignment(prg, tok, eFFStates::FogStart))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected float variable or literal in line %d", tok.getLine());
		}
		else if( tok.checkToken( "FogEnd" ) )
		{
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());
			if (!handleFloatAssignment(prg, tok, eFFStates::FogEnd))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected float variable or literal in line %d", tok.getLine());
		}
		else if( tok.checkToken( "FogColor" ) )
		{
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());
			if (!handleVec4Assignment(prg, tok, eFFStates::FogColor))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected vec4 variable or literal in line %d", tok.getLine());
		}
		else if( tok.checkToken( "LightAmbient" ) )
		{
			int index;
			if ((index=handleReadIndex(tok))<0)
				return prg->raiseError(GL_VERTEX_SHADER, "Expected index in form of '[<number>]' in line %d", tok.getLine());
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());

			int state = MAKE_INDEXED_STATE(eFFStates::LightAmbient, index);

			if (!handleVec4Assignment(prg, tok, state))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected vec4 variable or literal in line %d", tok.getLine());
		}
		else if( tok.checkToken( "LightDiffuse" ) )
		{
			int index;
			if ((index=handleReadIndex(tok))<0)
				return prg->raiseError(GL_VERTEX_SHADER, "Expected index in form of '[<number>]' in line %d", tok.getLine());
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());

			int state = MAKE_INDEXED_STATE(eFFStates::LightDiffuse, index);

			if (!handleVec4Assignment(prg, tok, state))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected vec4 variable or literal in line %d", tok.getLine());
		}
		else if( tok.checkToken( "LightSpecular" ) )
		{
			int index;
			if ((index=handleReadIndex(tok))<0)
				return prg->raiseError(GL_VERTEX_SHADER, "Expected index in form of '[<number>]' in line %d", tok.getLine());
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());

			int state = MAKE_INDEXED_STATE(eFFStates::LightSpecular, index);

			if (!handleVec4Assignment(prg, tok, state))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected vec4 variable or literal in line %d", tok.getLine());
		}
		else if( tok.checkToken( "LightEmission" ) )
		{
			int index;
			if ((index=handleReadIndex(tok))<0)
				return prg->raiseError(GL_VERTEX_SHADER, "Expected index in form of '[<number>]' in line %d", tok.getLine());
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());

			int state = MAKE_INDEXED_STATE(eFFStates::LightEmission, index);

			if (!handleVec4Assignment(prg, tok, state))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected vec4 variable or literal in line %d", tok.getLine());
		}
		else if( tok.checkToken( "LightPosition" ) )
		{
			int index;
			if ((index=handleReadIndex(tok))<0)
				return prg->raiseError(GL_VERTEX_SHADER, "Expected index in form of '[<number>]' in line %d", tok.getLine());
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());

			int state = MAKE_INDEXED_STATE(eFFStates::LightPosition, index);

			if (!handleVec4Assignment(prg, tok, state))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected vec4 variable or literal in line %d", tok.getLine());
		}
		else if( tok.checkToken( "LightSpotDirection" ) )
		{
			int index;
			if ((index=handleReadIndex(tok))<0)
				return prg->raiseError(GL_VERTEX_SHADER, "Expected index in form of '[<number>]' in line %d", tok.getLine());
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());

			int state = MAKE_INDEXED_STATE(eFFStates::LightSpotDirection, index);

			if (!handleVec3Assignment(prg, tok, state))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected vec3 variable or literal in line %d", tok.getLine());
		}
		else if( tok.checkToken( "LightSpotExponent" ) )
		{
			int index;
			if ((index=handleReadIndex(tok))<0)
				return prg->raiseError(GL_VERTEX_SHADER, "Expected index in form of '[<number>]' in line %d", tok.getLine());
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());

			int state = MAKE_INDEXED_STATE(eFFStates::LightSpotExponent, index);

			if (!handleFloatAssignment(prg, tok, state))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected float variable or literal in line %d", tok.getLine());
		}
		else if( tok.checkToken( "LightSpotCutoff" ) )
		{
			int index;
			if ((index=handleReadIndex(tok))<0)
				return prg->raiseError(GL_VERTEX_SHADER, "Expected index in form of '[<number>]' in line %d", tok.getLine());
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());

			int state = MAKE_INDEXED_STATE(eFFStates::LightSpotCutoff, index);

			if (!handleFloatAssignment(prg, tok, state))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected float variable or literal in line %d", tok.getLine());
		}
		else if( tok.checkToken( "LightConstantAttenuation" ) )
		{
			int index;
			if ((index=handleReadIndex(tok))<0)
				return prg->raiseError(GL_VERTEX_SHADER, "Expected index in form of '[<number>]' in line %d", tok.getLine());
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());

			int state = MAKE_INDEXED_STATE(eFFStates::LightConstantAttenuation, index);

			if (!handleFloatAssignment(prg, tok, state))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected float variable or literal in line %d", tok.getLine());
		}
		else if( tok.checkToken( "LightLinearAttenuation" ) )
		{
			int index;
			if ((index=handleReadIndex(tok))<0)
				return prg->raiseError(GL_VERTEX_SHADER, "Expected index in form of '[<number>]' in line %d", tok.getLine());
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());

			int state = MAKE_INDEXED_STATE(eFFStates::LightLinearAttenuation, index);

			if (!handleFloatAssignment(prg, tok, state))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected float variable or literal in line %d", tok.getLine());
		}
		else if( tok.checkToken( "LightQuadraticAttenuation" ) )
		{
			int index;
			if ((index=handleReadIndex(tok))<0)
				return prg->raiseError(GL_VERTEX_SHADER, "Expected index in form of '[<number>]' in line %d", tok.getLine());
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());

			int state = MAKE_INDEXED_STATE(eFFStates::LightQuadraticAttenuation, index);

			if (!handleFloatAssignment(prg, tok, state))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected float variable or literal in line %d", tok.getLine());
		}
		else if( tok.checkToken( "LightEnabled" ) )
		{
			int index;
			if ((index=handleReadIndex(tok))<0)
				return prg->raiseError(GL_VERTEX_SHADER, "Expected index in form of '[<number>]' in line %d", tok.getLine());
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());

			int state = MAKE_INDEXED_STATE(eFFStates::LightEnabled, index);

			int v = 0;
			if (tok.checkToken( "True")) v = 1;
			else if (tok.checkToken("False")) v = 0;
			else return prg->raiseError(GL_VERTEX_SHADER, "Expected True or False in line %d", tok.getLine());

			prg->mIntConstants.push_back(sIntConstant(state, v));
		}
		else if( tok.checkToken( "LightingEnabled" ) )
		{
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());

			int v = 0;
			if (tok.checkToken( "True")) v = 1;
			else if (tok.checkToken("False")) v = 0;
			else return prg->raiseError(GL_VERTEX_SHADER, "Expected True or False in line %d", tok.getLine());

			prg->mIntConstants.push_back(sIntConstant(eFFStates::LightingEnabled, v));
		}
		else if( tok.checkToken( "NormalizeEnabled" ) )
		{
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());

			int v = 0;
			if (tok.checkToken( "True")) v = 1;
			else if (tok.checkToken("False")) v = 0;
			else return prg->raiseError(GL_VERTEX_SHADER, "Expected True or False in line %d", tok.getLine());

			prg->mIntConstants.push_back(sIntConstant(eFFStates::NormalizeEnabled, v));
		}
		else if( tok.checkToken( "RescaleNormalEnabled" ) )
		{
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());

			int v = 0;
			if (tok.checkToken( "True")) v = 1;
			else if (tok.checkToken("False")) v = 0;
			else return prg->raiseError(GL_VERTEX_SHADER, "Expected True or False in line %d", tok.getLine());

			prg->mIntConstants.push_back(sIntConstant(eFFStates::RescaleNormalEnabled, v));
		}
		else if( tok.checkToken( "LightModelAmbient" ) )
		{
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());
			if (!handleVec4Assignment(prg, tok, eFFStates::LightModelAmbient))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected vec4 variable or literal in line %d", tok.getLine());
		}
		else if( tok.checkToken( "LightModelTwoSide" ) )
		{
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());

			int v = 0;
			if (tok.checkToken( "True")) v = 1;
			else if (tok.checkToken("False")) v = 0;
			else return prg->raiseError(GL_VERTEX_SHADER, "Expected True or False in line %d", tok.getLine());

			prg->mIntConstants.push_back(sIntConstant(eFFStates::LightModelTwoSide, v));
		}
		else if( tok.checkToken( "ModelViewMatrix" ) )
		{
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());
			if (!handleMat4Assignment(prg, tok, eFFStates::ModelViewMatrix))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected mat4 variable or literal in line %d", tok.getLine());
		}
		else if( tok.checkToken( "ProjectionMatrix" ) )
		{
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());
			if (!handleMat4Assignment(prg, tok, eFFStates::ProjectionMatrix))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected mat4 variable or literal in line %d", tok.getLine());
		}
		else if( tok.checkToken( "TextureMatrix" ) )
		{
			int index;
			if ((index=handleReadIndex(tok))<0)
				return prg->raiseError(GL_VERTEX_SHADER, "Expected index in form of '[<number>]' in line %d", tok.getLine());
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());
			
			int state = MAKE_INDEXED_STATE(eFFStates::TextureMatrix, index);

			if (!handleMat4Assignment(prg, tok, state))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected mat4 variable or literal in line %d", tok.getLine());
		}
		else if( tok.checkToken( "MaterialAmbient" ) )
		{
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());
			if (!handleVec4Assignment(prg, tok, eFFStates::MaterialAmbient))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected vec4 variable or literal in line %d", tok.getLine());
		}
		else if( tok.checkToken( "MaterialDiffuse" ) )
		{
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());
			if (!handleVec4Assignment(prg, tok, eFFStates::MaterialDiffuse))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected vec4 variable or literal in line %d", tok.getLine());
		}
		else if( tok.checkToken( "MaterialSpecular" ) )
		{
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());
			if (!handleVec4Assignment(prg, tok, eFFStates::MaterialSpecular))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected vec4 variable or literal in line %d", tok.getLine());
		}
		else if( tok.checkToken( "MaterialEmission" ) )
		{
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());
			if (!handleVec4Assignment(prg, tok, eFFStates::MaterialEmission))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected vec4 variable or literal in line %d", tok.getLine());
		}
		else if( tok.checkToken( "MaterialShininess" ) )
		{
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());
			if (!handleFloatAssignment(prg, tok, eFFStates::MaterialShininess))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected float variable or literal in line %d", tok.getLine());
		}
		else if( tok.checkToken( "ColorMaterialEnabled" ) )
		{
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());

			int v = 0;
			if (tok.checkToken( "True")) v = 1;
			else if (tok.checkToken("False")) v = 0;
			else return prg->raiseError(GL_VERTEX_SHADER, "Expected True or False in line %d", tok.getLine());

			prg->mIntConstants.push_back(sIntConstant(eFFStates::ColorMaterialEnabled, v));
		}
		else
			return prg->raiseError(GL_VERTEX_SHADER, "Invalid state: %s in line %d", tok.getToken(0), tok.getLine());

		if (!tok.checkToken( ";")) 
			return prg->raiseError(GL_VERTEX_SHADER, "Expected ';' in line %d", tok.getLine());
	}

	return GL_TRUE;
}

GLint glCompileFragmentShader(FixedFunctionProgram* prg, char* source)
{
	cTokenizer tok( source );
	const char *identifier = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_";
	const char *intnum = "+-0123456789";
	const char *floatnum = "+-0123456789.eE";

	while( tok.hasToken() )
	{
		if( tok.checkToken( "uniform" ) )
		{
			int uniformType = 0;
			if (tok.checkToken( "float") ) uniformType = GL_FLOAT;
			else if (tok.checkToken( "vec4") ) uniformType = GL_FLOAT_VEC4;
			else if (tok.checkToken( "sampler2D") ) uniformType = GL_SAMPLER_2D;
			else return prg->raiseError(GL_FRAGMENT_SHADER, "Invalid uniform type: %s in line %d", tok.getToken(0), tok.getLine());

			const char* id = tok.getToken( identifier );
			
			prg->mUniforms.push_back(sUniform(id,uniformType));
		}
		else if( tok.checkToken( "AlphaTestEnabled" ) )
		{
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());

			int v = 0;
			if (tok.checkToken( "True")) v = 1;
			else if (tok.checkToken("False")) v = 0;
			else return prg->raiseError(GL_VERTEX_SHADER, "Expected True or False in line %d", tok.getLine());

			prg->mIntConstants.push_back(sIntConstant(eFFStates::AlphaTestEnabled, v));
		}
		else if( tok.checkToken( "AlphaFunc" ) )
		{
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());

			int v = 0;
			if (tok.checkToken( "Never")) v = GL_NEVER;
			else if (tok.checkToken("Less")) v = GL_LESS;
			else if (tok.checkToken("Equal")) v = GL_EQUAL;
			else if (tok.checkToken("LEqual")) v = GL_LEQUAL;
			else if (tok.checkToken("Greater")) v = GL_GREATER;
			else if (tok.checkToken("NotEqual")) v = GL_NOTEQUAL;
			else if (tok.checkToken("GEqual")) v = GL_GEQUAL;
			else if (tok.checkToken("Always")) v = GL_ALWAYS;
			else return prg->raiseError(GL_VERTEX_SHADER, "Invalid comp function in line %d", tok.getLine());

			prg->mIntConstants.push_back(sIntConstant(eFFStates::AlphaFunc, v));
		}
		else if( tok.checkToken( "AlphaRef" ) )
		{
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());

			if (!handleFloatAssignment(prg, tok, eFFStates::AlphaRef))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected float variable or literal in line %d", tok.getLine());
		}
		else if( tok.checkToken( "Constant" ) )
		{
			int index;
			if ((index=handleReadIndex(tok))<0)
				return prg->raiseError(GL_VERTEX_SHADER, "Expected index in form of '[<number>]' in line %d", tok.getLine());
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected '=' in line %d", tok.getLine());

			int state = MAKE_INDEXED_STATE(eFFStates::Constant, index);

			if (!handleVec4Assignment(prg, tok, state))
				return prg->raiseError(GL_VERTEX_SHADER, "Expected vec4 variable or literal in line %d", tok.getLine());
		}
		else if( tok.checkToken( "Rgb", true ) || tok.checkToken( "Rgba", true ) || tok.checkToken( "Alpha" , true))
		{
			int type = 0;
			if (tok.checkToken("Rgb")) type = 0;
			else if (tok.checkToken("Rgba")) type = 1;
			else if (tok.checkToken("Alpha")) type = 2;

			//function - number of parameters - multiply
			int stateIndex;
			if ((stateIndex=handleReadIndex(tok))<0)
				return prg->raiseError(GL_FRAGMENT_SHADER, "Expected index in form of '[<number>]' in line %d", tok.getLine());
			if (!tok.checkToken( "="))
				return prg->raiseError(GL_FRAGMENT_SHADER, "Expected '=' in line %d", tok.getLine());

			int state = MAKE_INDEXED_STATE(type!=2?eFFStates::CombineRgb:eFFStates::CombineAlpha, stateIndex);

			int numParams = 0;
			if (tok.checkToken( "Replace") ) 
			{ 
				numParams = 1; 
				prg->mIntConstants.push_back(sIntConstant(state, GL_REPLACE));
			}
			else if (tok.checkToken( "Modulate") )
			{
				numParams = 2; 
				prg->mIntConstants.push_back(sIntConstant(state, GL_MODULATE));
			}
			else if (tok.checkToken( "Add") )
			{
				numParams = 2; 
				prg->mIntConstants.push_back(sIntConstant(state, GL_ADD));
			}
			else if (tok.checkToken( "AddSigned") )
			{
				numParams = 2; 
				prg->mIntConstants.push_back(sIntConstant(state, GL_ADD_SIGNED));
			}
			else if (tok.checkToken( "Interpolate") )
			{
				numParams = 3; 
				prg->mIntConstants.push_back(sIntConstant(state, GL_INTERPOLATE));
			}
			else if (tok.checkToken( "Subtract") )
			{
				numParams = 2; 
				prg->mIntConstants.push_back(sIntConstant(state, GL_SUBTRACT));
			}
			else if (tok.checkToken( "Dot3") )
			{
				numParams = 2; 
				int op = (type==1) ? GL_DOT3_RGBA : GL_DOT3_RGB;
				prg->mIntConstants.push_back(sIntConstant(state, op));
			}
			else 
				return prg->raiseError(GL_FRAGMENT_SHADER, "Invalid Operator: %s in line %d", tok.getToken(0), tok.getLine());

			const char* origWS = tok.getWhitespaceAndSpecial();
			tok.setWhitespaceAndSpecial(" \t\n\r{}()[]<>=,;.");	//hacking '.' as a special character to be able to tokenize

			if (!tok.checkToken( "(") ) 
				return prg->raiseError(GL_FRAGMENT_SHADER, "Expected '(' in line %d", tok.getLine());

			for(int i=0;i<numParams;++i)
			{
				tok.setWhitespaceAndSpecial(" \t\n\r{}()[]<>=,;.");

				if (i>0 && !tok.checkToken( ",") )
					return prg->raiseError(GL_FRAGMENT_SHADER, "Expected ',' in line %d", tok.getLine());

				int operand = 0;

				if (tok.checkToken("1") )
				{
					if (!tok.checkToken("-")) 
						return prg->raiseError(GL_FRAGMENT_SHADER, "Expected '-' after 1 in line %d", tok.getLine());
					operand |= 1;
				}

				int state = MAKE_INDEXED_STATE( (type!=2?eFFStates::Src0Rgb : eFFStates::Src0Alpha ) + i, stateIndex);


				if (tok.checkToken( "Texture") ) 
				{
/*					if (tok.checkToken( "["))
					{
						const char* texIdx = tok.getToken( intnum );
						int tIndex = atoi(texIdx);
						if (!tok.checkToken( "]")) 
							return prg->raiseError(GL_FRAGMENT_SHADER, "Expected ']' in line %d", tok.getLine());

						prg->mIntConstants.push_back(sIntConstant(state, GL_TEXTURE0+tIndex));
					}
					else*/
					{
						prg->mIntConstants.push_back(sIntConstant(state, GL_TEXTURE));
					}
				}
				else if (tok.checkToken( "Constant") )
				{
					prg->mIntConstants.push_back(sIntConstant(state, GL_CONSTANT));
				}
				else if (tok.checkToken( "PrimaryColor") )
				{
					prg->mIntConstants.push_back(sIntConstant(state, GL_PRIMARY_COLOR));
				}
				else if (tok.checkToken( "Previous") )
				{
					prg->mIntConstants.push_back(sIntConstant(state, GL_PREVIOUS));
				}
				else
					return prg->raiseError(GL_FRAGMENT_SHADER, "Invalid Operand:%s in line %d", tok.getToken(0), tok.getLine());


				if (tok.checkToken( ".") )
				{
					if (!tok.checkToken("A")) 
						return prg->raiseError(GL_FRAGMENT_SHADER, "Expected 'A' after '.' in line %d", tok.getLine());
					operand |= 2;
				}
				if ( type == 2) operand |= 2;

				tok.setWhitespaceAndSpecial(origWS);

				state = MAKE_INDEXED_STATE( (type!=2?eFFStates::Operand0Rgb : eFFStates::Operand0Alpha ) + i, stateIndex);
				if (operand == 0 ) operand = GL_SRC_COLOR;
				else if (operand == 1 ) operand = GL_ONE_MINUS_SRC_COLOR;
				else if (operand == 2 ) operand = GL_SRC_ALPHA;
				else if (operand == 3 ) operand = GL_ONE_MINUS_SRC_ALPHA;

				prg->mIntConstants.push_back(sIntConstant(state, operand));
			}

			if (!tok.checkToken( ")") )
				return prg->raiseError(GL_FRAGMENT_SHADER, "Expected ')' in line %d", tok.getLine());

			//read multiplier
			{
				float m = 1.0f;
				if (tok.checkToken( "*") )
				{
					const char* mult = tok.getToken( floatnum );
						//valid values are 1,2,4
					m = (float)atof(mult);
				}
				state = MAKE_INDEXED_STATE( type!=2?eFFStates::RgbScale : eFFStates::AlphaScale, stateIndex);
				prg->mFloatConstants.push_back(sFloatConstant(state, m));
			}

				//add texture2dEnabled
			{
				state = MAKE_INDEXED_STATE( eFFStates::Texture2DEnabled, stateIndex);
				bool found = false;
				for(unsigned int i=0;i<prg->mIntConstants.size();++i)
					if (prg->mIntConstants[i].mState == state)
					{
						found = true;
						break;
					}
				if (!found)
					prg->mIntConstants.push_back(sIntConstant(state, 1));
			}
		}
		else
			return prg->raiseError(GL_FRAGMENT_SHADER, "Invalid state: %s in line %d", tok.getToken(0), tok.getLine());

		if (!tok.checkToken( ";")) 
			return prg->raiseError(GL_FRAGMENT_SHADER, "Expected ';' in line %d", tok.getLine());
	}
	return 1;
}


void glGetShaderInfoLog(GLuint program, GLuint pname, GLsizei bufSize, GLsizei *length, GLchar *infoLog)
{	//get logs

	if (length)
		*length = 0;

	if (pname != GL_VERTEX_SHADER && pname != GL_FRAGMENT_SHADER)
		return;

	FixedFunctionProgram* p = gPrograms.getRef(program);
	if (p==NULL)
		return;

	const char* info = (pname == GL_VERTEX_SHADER) ? p->mVertexInfo.c_str() : p->mFragmentInfo.c_str();
	int s = (int)strlen(info)+1;
	if (bufSize < s) s = bufSize;
	memcpy(infoLog, info, s - 1);
	infoLog[s-1] = 0;

	if (length)
		*length = s;
}

void glGetProgramiv(GLuint program, GLenum pname, GLint *params)
{
	FixedFunctionProgram* p = gPrograms.getRef(program);
	if (p==NULL)
		return;

	switch(pname)
	{
	case GL_VERTEX_INFO_LOG_LENGTH:
		*params = (int)p->mVertexInfo.size()+1;
	break;
	case GL_FRAGMENT_INFO_LOG_LENGTH:
		*params = (int)p->mFragmentInfo.size()+1;
	break;
	case GL_ACTIVE_UNIFORMS:
		*params = (int)p->mUniforms.size();
	break;
	case GL_ACTIVE_UNIFORM_MAX_LENGTH:
	{
		size_t maxLength = 0;
		for(unsigned int i=0;i<p->mUniforms.size();++i)
			if (p->mUniforms[i].mName.size() > maxLength)
				maxLength = p->mUniforms[i].mName.size();
		*params = (int)maxLength;
	}
	break;
	case GL_ACTIVE_ATTRIBUTES:
	{
		*params = (int)p->mAttributes.size();
	}
	break;
	case GL_ACTIVE_ATTRIBUTE_MAX_LENGTH:
	{
		size_t maxLength = 0;
		for(unsigned int i=0;i<p->mAttributes.size();++i)
			if (p->mAttributes[i].mSemantic.size() > maxLength)
				maxLength = p->mAttributes[i].mSemantic.size();
		*params = (int)maxLength;
	}
	break;
	}
}


void glGetActiveAttrib(GLuint program, GLuint index, GLenum* attr, GLenum *type, 
							GLsizei bufSize, GLsizei *length, GLchar *semantic)
{
	FixedFunctionProgram* p = gPrograms.getRef(program);
	if (p==NULL)
		return;

	sAttribute& attrib = p->mAttributes[index];

	*attr = attrib.mID;
	*type = attrib.mType;
	int s = (int)attrib.mSemantic.size()+1;
	if (bufSize < s) s = bufSize;
	if (length) *length = s;
	memcpy(semantic, attrib.mSemantic.c_str(), s - 1);
	semantic[s-1] = 0;
}

GLint glGetUniformLocation(GLuint program, const GLchar *name)
{
	FixedFunctionProgram* p = gPrograms.getRef(gCurrentProgram);
	if (p==NULL)
		return -1;

	for(unsigned int i=0;i<p->mUniforms.size();++i)
		if (strcmp(name, p->mUniforms[i].mName.c_str())==0)
			return i;

	return -1;
}

void glGetActiveUniform(GLuint program, GLuint index, GLsizei bufSize, GLsizei *length,
 	GLenum *type, GLchar *name)
{
	if (length)
		*length = 0;

	FixedFunctionProgram* p = gPrograms.getRef(gCurrentProgram);
	if (p==NULL)
		return;

	sUniform& u = p->mUniforms[index];

	if (type)
		*type = u.mType;

	int s = (int)u.mName.size()+1;
	if (bufSize < s) s = bufSize;
	if (length) *length = s;
	memcpy(name, u.mName.c_str(), s - 1);
	name[s-1] = 0;
}

void glUniform1f(GLint location, GLfloat v0)
{
	FixedFunctionProgram* p = gPrograms.getRef(gCurrentProgram);
	if (p==NULL)
		return;

	sUniform& u = p->mUniforms[location];

	for(unsigned int i=0;i<u.mStates.size();++i)
	{
		int state = u.mStates[i];
		switch(GET_STATE(state))
		{
		case eFFStates::FogDensity:
			glFogf(GL_FOG_DENSITY, v0);
		break;
		case eFFStates::FogStart:
			glFogf(GL_FOG_START, v0);
		break;
		case eFFStates::FogEnd:
			glFogf(GL_FOG_END, v0);
		break;
		case eFFStates::FogMode:
			glFogf(GL_FOG_MODE, v0);
		break;

		case eFFStates::LightModelTwoSide:
			glLightModelf(GL_LIGHT_MODEL_TWO_SIDE, v0);
		break;

		case eFFStates::LightSpotExponent:
			glLightf(GL_LIGHT0 + GET_INDEX(state), GL_SPOT_EXPONENT, v0);
		break;
		case eFFStates::LightSpotCutoff:
			glLightf(GL_LIGHT0 + GET_INDEX(state), GL_SPOT_CUTOFF, v0);
		break;
		case eFFStates::LightConstantAttenuation:
			glLightf(GL_LIGHT0 + GET_INDEX(state), GL_CONSTANT_ATTENUATION, v0);
		break;
		case eFFStates::LightLinearAttenuation:
			glLightf(GL_LIGHT0 + GET_INDEX(state), GL_LINEAR_ATTENUATION, v0);
		break;
		case eFFStates::LightQuadraticAttenuation:
			glLightf(GL_LIGHT0 + GET_INDEX(state), GL_QUADRATIC_ATTENUATION, v0);
		break;
		case eFFStates::MaterialShininess:
			glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, v0);
		break;
		case eFFStates::AlphaRef:
		{
			int func;
			glGetIntegerv(GL_ALPHA_TEST_FUNC, &func);
			glAlphaFunc(func, v0);
		}
		break;
		case eFFStates::RgbScale:
			glTexEnvf(GL_TEXTURE_ENV, GL_RGB_SCALE, v0);
		break;
		case eFFStates::AlphaScale:
			glTexEnvf(GL_TEXTURE_ENV, GL_ALPHA_SCALE, v0);
		break;
		default:
			;
			//invalid state
		}
	}
}

void glUniform3f(GLint location, GLfloat v0, GLfloat v1, GLfloat v2)
{
	FixedFunctionProgram* p = gPrograms.getRef(gCurrentProgram);
	if (p==NULL)
		return;

	sUniform& u = p->mUniforms[location];

	float v[3] = {v0,v1,v2};

	for(unsigned int i=0;i<u.mStates.size();++i)
	{
		int state = u.mStates[i];
		switch(GET_STATE(state))
		{
		case eFFStates::LightSpotDirection:
			glLightfv(GL_LIGHT0 + GET_INDEX(state), GL_SPOT_DIRECTION, v);
		break;
		default:
			;
			//invalid state
		}
	}
}

void glUniform4f(GLint location, GLfloat v0, GLfloat v1, GLfloat v2, GLfloat v3)
{
	FixedFunctionProgram* p = gPrograms.getRef(gCurrentProgram);
	if (p==NULL)
		return;

	sUniform& u = p->mUniforms[location];
	float v[4] = {v0,v1,v2,v3};

	for(unsigned int i=0;i<u.mStates.size();++i)
	{
		int state = u.mStates[i];
		switch(GET_STATE(state))
		{
		case eFFStates::Color:	// vec4
			glColor4f(v0,v1,v2,v3);
		break;
		case eFFStates::FogColor: // vec4
			glFogfv(GL_FOG_COLOR, v);
		break;
		case eFFStates::LightModelAmbient: // vec4
			glLightModelfv(GL_LIGHT_MODEL_AMBIENT , v);
		break;
		case eFFStates::LightAmbient:
			glLightfv(GL_LIGHT0 + GET_INDEX(state), GL_AMBIENT, v);
		break;
		case eFFStates::LightDiffuse:
			glLightfv(GL_LIGHT0 + GET_INDEX(state), GL_DIFFUSE, v);
		break;
		case eFFStates::LightSpecular:
			glLightfv(GL_LIGHT0 + GET_INDEX(state), GL_SPECULAR, v);
		break;
		case eFFStates::LightEmission:
			glLightfv(GL_LIGHT0 + GET_INDEX(state), GL_EMISSION, v);
		break;
		case eFFStates::LightPosition:
			glLightfv(GL_LIGHT0 + GET_INDEX(state), GL_POSITION, v);
		break;
		case eFFStates::MaterialAmbient:
			glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, v);
		break;
		case eFFStates::MaterialDiffuse:
			glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, v);
		break;
		case eFFStates::MaterialSpecular: // vec4
			glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, v);
		break;
		case eFFStates::MaterialEmission: //  vec4
			glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, v);
		break;

		// Fragment pipeline
		case eFFStates::Constant: // [n] = vec4Variable;
		{
			glActiveTexture(GL_TEXTURE0 + GET_INDEX(state));
			glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, v);
		}
		break;
		}
	}
}

void glUniformMatrix4fv(GLint location, GLsizei count, GLboolean transpose, const GLfloat *value)
{
	FixedFunctionProgram* p = gPrograms.getRef(gCurrentProgram);
	if (p==NULL)
		return;

	ASSERT(count == 1 && transpose == false);

	sUniform& u = p->mUniforms[location];

	for(unsigned int i=0;i<u.mStates.size();++i)
	{
		int state = u.mStates[i];
		switch(GET_STATE(state))
		{
		case eFFStates::ModelViewMatrix:
			glMatrixMode(GL_MODELVIEW);
			glLoadMatrixf(value);
		break;
		case eFFStates::ProjectionMatrix:
			glMatrixMode(GL_PROJECTION);
			glLoadMatrixf(value);
		break;
		case eFFStates::TextureMatrix:	
			glActiveTexture(GL_TEXTURE0 + GET_INDEX(state));//active texture unit
			glMatrixMode(GL_TEXTURE);
			glLoadMatrixf(value);
		break;
		}
	}
}

bool handleFloatAssignment(FixedFunctionProgram* prg, cTokenizer& tok, int state)
{
	const char *identifier = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_";
	const char *floatnum = "+-0123456789.eE";

	const char* id = tok.getToken( identifier );

	int uniformIndex = -1;

	for(unsigned int i=0;i<prg->mUniforms.size();++i)
		if (strcmp(id, prg->mUniforms[i].mName.c_str())==0)
		{
			uniformIndex = i;
			break;
		}

	if	(uniformIndex != -1)
	{
		if (prg->mUniforms[uniformIndex].mType != GL_FLOAT)
			return false;

		prg->mUniforms[uniformIndex].mStates.push_back(state);
	}
	else
	{	//try to read as float
		const char* fs = tok.getToken( floatnum );	
		if (*fs==0) return false;
		float v = (float)atof(fs);

		//add constant
		prg->mFloatConstants.push_back(sFloatConstant(state, v));
	}

	return true;
}



bool handleVec3Assignment(FixedFunctionProgram* prg, cTokenizer& tok, int state)
{
	const char *identifier = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_";
	const char *floatnum = "+-0123456789.eE";

	if (tok.checkToken( "vec3"))
	{
		float v[3];

		if (!tok.checkToken("(")) return false;
		for(unsigned int i=0;i<3;++i)
		{
			const char* fs = tok.getToken( floatnum );
			v[i] = (float)atof(fs);
			if (i<2 && !tok.checkToken( ",")) return false;
		}

		if (!tok.checkToken( ")")) return false;

		//add constant
		prg->mVec3Constants.push_back(sVec3Constant(state, v));
	}
	else
	{
		const char* id = tok.getToken( identifier );

		int uniformIndex = -1;

		for(unsigned int i=0;i<prg->mUniforms.size();++i)
			if (strcmp(id, prg->mUniforms[i].mName.c_str())==0)
			{
				uniformIndex = i;
				break;
			}

		if	(uniformIndex == -1)
			return false;

		if (prg->mUniforms[uniformIndex].mType != GL_FLOAT_VEC3)
			return false;

		prg->mUniforms[uniformIndex].mStates.push_back(state);
	}

	return true;
}


bool handleVec4Assignment(FixedFunctionProgram* prg, cTokenizer& tok, int state)
{
	const char *identifier = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_";
	const char *floatnum = "+-0123456789.eE";

	if (tok.checkToken( "vec4"))
	{
		float v[4];

		if (!tok.checkToken("(")) return false;
		for(unsigned int i=0;i<4;++i)
		{
			const char* fs = tok.getToken( floatnum );
			v[i] = (float)atof(fs);
			if (i<3 && !tok.checkToken( ",")) return false;
		}

		if (!tok.checkToken( ")")) return false;

		//add constant
		prg->mVec4Constants.push_back(sVec4Constant(state, v));
	}
	else
	{
		const char* id = tok.getToken( identifier );

		int uniformIndex = -1;

		for(unsigned int i=0;i<prg->mUniforms.size();++i)
			if (strcmp(id, prg->mUniforms[i].mName.c_str())==0)
			{
				uniformIndex = i;
				break;
			}

		if	(uniformIndex == -1)
			return false;

		if (prg->mUniforms[uniformIndex].mType != GL_FLOAT_VEC4)
			return false;

		prg->mUniforms[uniformIndex].mStates.push_back(state);
	}

	return true;
}

bool handleMat4Assignment(FixedFunctionProgram* prg, cTokenizer& tok, int state)
{
	const char *identifier = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_";
	const char *floatnum = "+-0123456789.eE";

	if (tok.checkToken( "mat4"))
	{
		float v[16];

		if (!tok.checkToken("(")) return false;
		for(unsigned int i=0;i<16;++i)
		{
			const char* fs = tok.getToken( floatnum );
			v[i] = (float)atof(fs);
			if (i<15 && !tok.checkToken( ",")) return false;
		}

		if (!tok.checkToken( ")")) return false;

		//add constant
		prg->mMat4Constants.push_back(sMat4Constant(state, v));
	}
	else
	{
		const char* id = tok.getToken( identifier );

		int uniformIndex = -1;

		for(unsigned int i=0;i<prg->mUniforms.size();++i)
			if (strcmp(id, prg->mUniforms[i].mName.c_str())==0)
			{
				uniformIndex = i;
				break;
			}

		if	(uniformIndex == -1)
			return false;

		if (prg->mUniforms[uniformIndex].mType != GL_FLOAT_MAT4)
			return false;

		prg->mUniforms[uniformIndex].mStates.push_back(state);
	}

	return true;
}


int handleReadIndex(cTokenizer& tok)
{
	const char *intnum = "+0123456789";
	if (!tok.checkToken( "[")) return -1;
	const char* idx = tok.getToken( intnum );
	if (*idx==0) return -1;
	if (!tok.checkToken( "]")) return -1;
	return atoi(idx);
}


void removeComments(char *text)
{
	char *p = text;
	while( *p )
	{
		if( *p == '/' && *(p+1) == '/' )
		{
			while( *p && *p != '\n' && *p != '\r' )
				*p++ = ' ';
			if (p==0) break;
		}
		else if( *p == '/' && *(p+1) == '*' )
		{
			*p++ = ' '; *p++ = ' ';
			if (*p==0) break;
			while( *(p+1) && (*p != '*' || *(p+1) != '/') )
				if ( *p != '\n' && *p != '\r')
					*p++ = ' ';
				else
					++p;
			*p++ = ' '; *p++ = ' ';
		}
		++p;
	}
}

void preprocessMacro(char *text)
{
	std::vector<std::string> defines;

	std::vector<bool>	defResult;

	bool replaceLineWithSpaces = false;

	char *p = text;
	while( *p )
	{
		//get first non-whitespace character
		char* lineStart = p;

		while (*p == ' ' || *p == '\t')
			++p;

		//if # we have to process this
		if (*p == '#')
		{
			++p;
			replaceLineWithSpaces = true;

			const char *identifier = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_";
			char* lineStart = p;
				//go to end of line
			while (*p && *p != '\r' && *p != '\n')
					++p;

			char lastChar = *p;
			*p = 0;


			cTokenizer tok(lineStart);

				//first token
			if (tok.checkToken("define"))
			{
				const char* d = tok.getToken(identifier);
				if (*d)
					defines.push_back(d);
			}
			else if (tok.checkToken("undef"))
			{
				const char* d = tok.getToken(identifier);
				if (*d)
				{
					for(unsigned int i=0;i<defines.size();++i)
						if (!defines[i].compare(d))
						{
							defines.erase(defines.begin()+i);
							break;
						}
				}
			}
			else if (tok.checkToken("ifdef"))
			{
				const char* d = tok.getToken(identifier);
				if (*d)
				{
					bool found = false;
					for(unsigned int i=0;i<defines.size();++i)
						if (!defines[i].compare(d))
						{
							found = true;
							break;
						}
					defResult.push_back(found);
				}
			}
			else if (tok.checkToken("ifndef"))
			{
				const char* d = tok.getToken(identifier);
				if (*d)
				{
					bool found = false;
					for(unsigned int i=0;i<defines.size();++i)
						if (!defines[i].compare(d))
						{
							found = true;
							break;
						}
					defResult.push_back(!found);
				}
			}
			else if (tok.checkToken("else"))
			{
				if (defResult.size())
					defResult[defResult.size()-1] = !defResult[defResult.size()-1];
			}
			else if (tok.checkToken("endif"))
			{
				if (defResult.size())
					defResult.pop_back();
			}

			*p = lastChar;
		}
		else
		{
			replaceLineWithSpaces = false;

				//go to end of line
			while (*p && *p != '\r' && *p != '\n')
					++p;
		}

		if (!replaceLineWithSpaces)
		{
			for(unsigned int i=0;i<defResult.size();++i)
				if (!defResult[i])
				{
					replaceLineWithSpaces = true;
					break;
				}
		}

		if (replaceLineWithSpaces)
			memset(lineStart, ' ', p-lineStart);

			//go to next line
		while (*p && (*p == '\r' || *p == '\n'))
			++p;
	}

}
